
// Asounding Audio Engine for iOS
// Copyright (c) 2011  Michel Fortin
//
// You must obtain a comercial license for using this code in software
// distributed not under the GNU GPL. See <http://michelf.com/asounding/>.
//
// This program is free software: you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation, either version 3 of the License, or (at your option) 
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program. If not, see <http://www.gnu.org/licenses/>.
//
// As a special exception to the GPL, you may distribute a compiled 
// application including this file on the Apple App Store for iOS without 
// having to redistribute the digital keys needed to reproduce the digital 
// signature needed to run an application on an iOS device. This exception 
// does not invalidate any other requirements put on you by the GPL.

#import "ASNEngine.h"
#import "ASNMusic.h"
#import "ASNSource.h"
#import "ASNOpenALSupport.h"
#import <AVFoundation/AVFoundation.h>
#import <AudioToolbox/AudioToolbox.h>
#import <UIKit/UIKit.h> // for application notifications
#import <unistd.h> // for usleep

@interface ASNEngine ( AsoundingPrivate )

- (void)initOpenAL;
- (void)teardownOpenAL;

- (void)startInterruption;
- (void)stopInterruption;

- (void)updateAudioState;

@end

#define MUSIC_CONTROL_NOOP -1
#define MUSIC_CONTROL_DONE 0
#define MUSIC_CONTROL_PLAY 1
#define MUSIC_CONTROL_PAUSE 2
#define MUSIC_CONTROL_STOP 3
#define MUSIC_CONTROL_FADE 4

#define MUSIC_FADE_QUANTUM 0.08


void interruptionListener(	void *	inClientData,
						  UInt32	inInterruptionState)
{
	ASNEngine* engine = (ASNEngine*)inClientData;
	if (inInterruptionState == kAudioSessionBeginInterruption)
	{		
		[engine startInterruption];
	}
	else if (inInterruptionState == kAudioSessionEndInterruption)
	{
		[engine stopInterruption];
	}
}

void RouteChangeListener(void *                  inClientData,
						 AudioSessionPropertyID	inID,
						 UInt32                  inDataSize,
						 const void *            inData)
{
	CFDictionaryRef dict = (CFDictionaryRef)inData;
	
	CFStringRef oldRoute = CFDictionaryGetValue(dict, CFSTR(kAudioSession_AudioRouteChangeKey_OldRoute));
	
	UInt32 size = sizeof(CFStringRef);
	
	CFStringRef newRoute;
	OSStatus result = AudioSessionGetProperty(kAudioSessionProperty_AudioRoute, &size, &newRoute);
	
	NSLog(@"result: %d Route changed from %@ to %@", result, oldRoute, newRoute);
}


@implementation ASNEngine


static ASNEngine *sharedInstance = nil;

+ (ASNEngine *)sharedEngine
{
	if (!sharedInstance) {
		@synchronized (self) {
			if (!sharedInstance)
				sharedInstance = [[self alloc] init];
		}
	}
	return sharedInstance;	
}

- (id)init
{	
	if (self = [super init]) {
		sources = [[NSMutableSet alloc] init];
		musicVolume = 0.5;
		isAudioPlaybackSuspended = YES;
		
		// make sure control thread exists
		musicControlState = MUSIC_CONTROL_NOOP;
		musicControlCondition = [[NSCondition alloc] init];
		musicControlThread = [[NSThread alloc] initWithTarget:self selector:@selector(runMusicControlThread:) object:nil];
		[musicControlThread start];
		
		// setup our audio session
		OSStatus result = AudioSessionInitialize(NULL, NULL, interruptionListener, self);
		if (result) NSLog(@"Error initializing audio session! %d\n", result);
		
		isAudioSessionInterrupted = NO;
		isApplicationActive = YES;
		isMusicEnabled = YES;
		isSoundsEnabled = YES;
		isAudioEnabled = YES;
		iPodPlaybackDisabled = NO;
		
		[self updateAudioState];
		
		// Initialize our OpenAL environment
		[self initOpenAL];
		
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidBecomeActive:) name:UIApplicationDidBecomeActiveNotification object:nil];
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillResignActive:) name:UIApplicationWillResignActiveNotification object:nil];
	}
	
	return self;
}

- (void)dealloc
{	
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	[self teardownOpenAL];
	[music release];
	[sources release];
	
	musicControlState = MUSIC_CONTROL_DONE;
	[musicControlCondition signal];
	// FIXME: find a better way to wait for the thread to finish.
	while ([musicControlThread isExecuting])
		usleep(1);
	[musicControlThread release];
	[musicControlCondition release];
	
	[super dealloc];
}


@dynamic audioSessionInterrupted;
@dynamic iPodIsPlaying;

- (BOOL)audioSessionInterrupted {
	return isAudioSessionInterrupted;
}

- (void)startInterruption {
	[musicControlCondition lock];
	isAudioSessionInterrupted == NO;
	[self updateAudioState];
	[musicControlCondition unlock];
}

- (void)stopInterruption {
	[musicControlCondition lock];
	isAudioSessionInterrupted == YES;
	[self updateAudioState];
	[musicControlCondition unlock];
}

- (void)applicationDidBecomeActive:(NSNotification *)notification {
	[musicControlCondition lock];
	isApplicationActive = YES;
	[self updateAudioState];
	[musicControlCondition unlock];
}

- (void)applicationWillResignActive:(NSNotification *)notification {
	[musicControlCondition lock];
	isApplicationActive = NO;
	[self updateAudioState];
	[musicControlCondition unlock];
}

@dynamic audioPlaybackSuspended;

- (void)activateAudioSessionSolo:(BOOL)solo {
	OSStatus result;
	
	if (solo) {
		//NSLog(@"Setting audio session no external audio.");
		UInt32 category = kAudioSessionCategory_SoloAmbientSound;
		result = AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(category), &category);
		if (result) NSLog(@"Asounding: Error setting solo audio session category: %@ (Note: this error is normal when running in the simulator.)\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
		
		AudioSessionSetActive(YES);
		if (result) NSLog(@"Asounding: Error setting audio session active: %@ (Note: this error is normal when running in the simulator.)\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
		
		category = kAudioSessionCategory_AmbientSound;
		result = AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(category), &category);
		if (result) NSLog(@"Asounding: Error setting audio session category: %@\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
	} else {
		//NSLog(@"Setting audio session with external audio.");
		UInt32 category = kAudioSessionCategory_AmbientSound;
		result = AudioSessionSetProperty(kAudioSessionProperty_AudioCategory, sizeof(category), &category);
		if (result) NSLog(@"Asounding: Error setting audio session category: %@ (Note: this error is normal when running in the simulator.)\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
		
		AudioSessionSetActive(YES);
		if (result) NSLog(@"Asounding: Error setting audio session active: %@ (Note: this error is normal when running in the simulator.)\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
	}
}

- (void)deactivateAudioSession {
	OSStatus result = AudioSessionSetActive(NO);
	if (result) NSLog(@"Asounding: Error setting audio session inactive: %@\n", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
}

- (void)updateAudioState {
	{
		BOOL newAudioSuspended = isAudioSessionInterrupted || !isAudioEnabled || !isApplicationActive;
		if (isAudioPlaybackSuspended == newAudioSuspended)
			return;
		isAudioPlaybackSuspended = newAudioSuspended;
	}
	
	if (isAudioPlaybackSuspended)
	{
		alcMakeContextCurrent(NULL);
		[[music avPlayer] stop];
		[self deactivateAudioSession];
	}
	else
	{
		if (iPodPlaybackDisabled)
			iPodIsPlaying = NO;
		else
		{
			// check for other audio playing, in which case we don't want to 
			// play the background music
			UInt32 size = sizeof(iPodIsPlaying);
			OSStatus result = AudioSessionGetProperty(kAudioSessionProperty_OtherAudioIsPlaying, &size, &iPodIsPlaying);
			if (result) NSLog(@"Error getting other audio playing property: %@", [[NSError errorWithDomain:(NSString *)kCFErrorDomainOSStatus code:result userInfo:nil] localizedDescription]);
		}
		
		[self activateAudioSessionSolo:!iPodIsPlaying];
		
		[musicControlCondition signal];
		
		alcMakeContextCurrent(context);
		if (!isSoundsEnabled)
			[sources makeObjectsPerformSelector:@selector(stop)];
	}
}


@dynamic audioEnabled;

- (BOOL)isAudioEnabled {
	return isAudioEnabled;
}

- (void)setAudioEnabled:(BOOL)flag {
	if (flag != isAudioEnabled) {
		[musicControlCondition lock];
		isAudioEnabled = flag;
		[self updateAudioState];
		[musicControlCondition signal];
		[musicControlCondition unlock];
	}
}

@dynamic music;

- (ASNMusic *)music {
	return music;
}

- (void)setMusic:(ASNMusic *)newMusic {
	[[music avPlayer] stop];
	[music release];
	music = [newMusic retain];
	[music avPlayer].currentTime = 0;
}


@dynamic musicVolume;

- (float)musicVolume {
	return musicVolume;
}

- (void)setMusicVolume:(float)newVolume {
	musicVolume = newVolume;
	[music avPlayer].volume = musicVolume;
}


@dynamic musicEnabled;

- (BOOL)isMusicEnabled {
	return isMusicEnabled;
}

- (void)setMusicEnabled:(BOOL)flag {
	if (flag != isMusicEnabled) {
		[musicControlCondition lock];
		isMusicEnabled = flag;
		[musicControlCondition signal];
		[musicControlCondition unlock];
	}
}

@dynamic iPodPlaybackDisabled;

- (BOOL)isIPodPlaybackDisabled {
	return iPodPlaybackDisabled;
}

- (void)setIPodPlaybackDisabled:(BOOL)flag {
	if (iPodPlaybackDisabled != flag) {
		if (iPodIsPlaying) {
			// Stop iPod playback.
			alcMakeContextCurrent(nil);
			[self deactivateAudioSession];
			[self activateAudioSessionSolo:flag];
			alcMakeContextCurrent(context);
			iPodIsPlaying = NO;
		}
		[musicControlCondition lock];
		iPodPlaybackDisabled = flag;
		[musicControlCondition signal];
		[musicControlCondition unlock];
	}
}

@dynamic musicPlaying;

- (BOOL)isMusicPlaying {
	[musicControlCondition lock];
	BOOL isPlaying = musicControlState == MUSIC_CONTROL_PLAY;
	[musicControlCondition unlock];
	return isPlaying;
}


@dynamic musicPaused;

- (BOOL)isMusicPaused {
	[musicControlCondition lock];
	BOOL isPaused = musicControlState == MUSIC_CONTROL_PAUSE;
	[musicControlCondition unlock];
	return isPaused;
}


- (BOOL)shouldPlayMusic {
	return isMusicEnabled && !isAudioPlaybackSuspended && !iPodIsPlaying;
}

- (void)runMusicControlThread:(id)nothingness {
	[musicControlCondition lock];
	
	while (musicControlState != MUSIC_CONTROL_DONE) {
		NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
		if (musicControlState == MUSIC_CONTROL_FADE)
			[musicControlCondition waitUntilDate:[NSDate dateWithTimeIntervalSinceNow:MUSIC_FADE_QUANTUM]];
		else
			[musicControlCondition wait];
		
		switch (musicControlState) {
			case MUSIC_CONTROL_PLAY:
				if ([self shouldPlayMusic]) {
					[music avPlayer].volume = musicVolume;
					[[music avPlayer] play];
				} else {
					[[music avPlayer] pause];
				}
				break;
			case MUSIC_CONTROL_PAUSE:
				[[music avPlayer] pause];
				break;
			case MUSIC_CONTROL_STOP:
				[[music avPlayer] stop];
				break;
			case MUSIC_CONTROL_FADE:
			{
				float newVolume = [music avPlayer].volume - musicFadeStep;
				if (newVolume > 0) {
					[music avPlayer].volume = newVolume;
				}
				else {
					musicControlState = MUSIC_CONTROL_STOP;
					[[music avPlayer] stop];
				}
				break;
			}
		}
		[pool release];
	}
		
	[musicControlCondition unlock];
}

- (void)terminateMusicControl {
	[musicControlCondition lock];
	musicControlState = MUSIC_CONTROL_DONE;
	[musicControlCondition signal];
	[musicControlCondition unlock];
}

- (void)playMusic {
	[musicControlCondition lock];
	
	musicControlState = MUSIC_CONTROL_PLAY;
	[musicControlCondition signal];
	
	[musicControlCondition unlock];
}

- (void)pauseMusic {
	[musicControlCondition lock];
	musicControlState = MUSIC_CONTROL_PAUSE;
	[[music avPlayer] pause];
	[musicControlCondition unlock];
}

- (void)stopMusic {
	[musicControlCondition lock];
	musicControlState = MUSIC_CONTROL_STOP;
	[[music avPlayer] stop];
	[musicControlCondition unlock];
}

- (void)fadeMusicWithin:(NSTimeInterval)seconds {
	if (musicControlState == MUSIC_CONTROL_PLAY)
	{
		[musicControlCondition lock];
		musicControlState = MUSIC_CONTROL_FADE;
		musicFadeStep = musicVolume / (seconds / MUSIC_FADE_QUANTUM);
		[musicControlCondition signal];
		[musicControlCondition unlock];
	}
}


@dynamic soundsEnabled;

- (BOOL)isSoundsEnabled {
	return isSoundsEnabled;
}

- (void)setSoundsEnabled:(BOOL)flag {
	if (flag != isSoundsEnabled) {
		isSoundsEnabled = flag;
		[sources makeObjectsPerformSelector:@selector(setEngine:) 
								 withObject:isSoundsEnabled ? self : nil];
		if (!isAudioPlaybackSuspended)
			[sources makeObjectsPerformSelector:@selector(stop)];
	}
}

@dynamic sources;

- (NSSet *)sources {
	return [sources copy];
}

- (void)addSource:(ASNSource *)newSource {
	[sources addObject:newSource];
	if (isSoundsEnabled)
		newSource.engine = self;
}

- (void)removeSource:(ASNSource *)theSource {
	theSource.engine = nil;
	[sources removeObject:theSource];
}


#pragma mark OpenAL


- (void)initOpenAL
{
	ALenum			error;
	
	// Create a new OpenAL Device
	// Pass NULL to specify the system’s default output device
	device = alcOpenDevice(NULL);
	if (device != NULL)
	{
		// Create a new OpenAL Context
		// The new context will render to the OpenAL Device just created 
		context = alcCreateContext(device, 0);
		if (context != NULL)
		{
			// Make the new context the Current OpenAL Context
			alcMakeContextCurrent(context);
		}
	}
	// clear any errors
	error = alGetError();
}

- (void)teardownOpenAL
{	
	// Delete the Sources
//    alDeleteSources(1, &source);
	// Delete the Buffers
//    alDeleteBuffers(1, &buffer);
	
    //Release context
    alcDestroyContext(context);
    //Close device
    alcCloseDevice(device);
}

@end
